package ip;

import addr.addrIP;
import addr.addrPrefix;
import cfg.cfgAll;
import cfg.cfgRtr;
import cfg.cfgVrf;
import clnt.clntMplsTeP2p;
import ifc.ifcNull;
import java.util.ArrayList;
import pack.packHolder;
import pack.packRsvp;
import rtr.rtrLdpNeigh;
import tab.tabHop;
import tab.tabLabel;
import tab.tabLabelNtry;
import tab.tabListing;
import tab.tabNatCfgN;
import tab.tabNatTraN;
import tab.tabPrfxlstN;
import tab.tabRoute;
import tab.tabRouteEntry;
import user.userFormat;
import util.bits;
import util.debugger;
import util.logger;

/**
 * calculates ip forwarding tables
 *
 * @author matecsaba
 */
public class ipFwdTab {

    /**
     * convert interface number to name
     *
     * @param fwd forwarding core to use
     * @param iface interface number
     * @return interface name
     */
    public static String iface2name(ipFwd fwd, int iface) {
        if (iface == 0) {
            return "null";
        }
        if (fwd == null) {
            return "#" + iface;
        }
        ipFwdIface ifc = fwd.ifaces.find(new ipFwdIface(iface, null));
        if (ifc == null) {
            return "!" + iface;
        }
        return "" + ifc;
    }

    /**
     * find originating interface to address
     *
     * @param lower forwarder
     * @param adr address to look to
     * @return interface to use
     */
    public static ipFwdIface findSendingIface(ipFwd lower, addrIP adr) {
        tabRouteEntry<addrIP> prf = lower.actualR.route(adr);
        if (prf == null) {
            return null;
        }
        if (prf.rouTab != null) {
            return findStableIface(lower);
        }
        return (ipFwdIface) prf.iface;
    }

    /**
     * find connected interface to address
     *
     * @param lower forwarder
     * @param adr address to look to
     * @return interface to use
     */
    public static ipFwdIface findConnedIface(ipFwd lower, addrIP adr) {
        tabRouteEntry<addrIP> prf = lower.actualR.route(adr);
        if (prf == null) {
            return null;
        }
        switch (prf.rouTyp) {
            case conn:
            case remote:
            case defpref:
            case automesh:
                break;
            default:
                return null;
        }
        return (ipFwdIface) prf.iface;
    }

    /**
     * find stable interface
     *
     * @param lower forwarder
     * @return interface id, null if none
     */
    public static ipFwdIface findStableIface(ipFwd lower) {
        ipFwdIface best = new ipFwdIface(0, null);
        for (int i = 0; i < lower.ifaces.size(); i++) {
            ipFwdIface ifc = lower.ifaces.get(i);
            if (ifc.mask > best.mask) {
                best = ifc;
            }
        }
        if (best.mask < 1) {
            return null;
        }
        return best;
    }

    /**
     * list protocols
     *
     * @param lower forwarder
     * @param l list to update
     * @param b beginning to use
     */
    public static void listProtocols(ipFwd lower, userFormat l, String b) {
        lower.protos.dump(lower, l, b);
    }

    /**
     * list ldp neighbors
     *
     * @param lower forwarder
     * @return list of neighbors
     */
    public static userFormat ldpNeighShow(ipFwd lower) {
        userFormat l = new userFormat("|", "learn|advert|l2learn|l2advert|mplearn|mpadvert|neighbor|uptime");
        for (int i = 0; i < lower.ldpNeighs.size(); i++) {
            rtrLdpNeigh ntry = lower.ldpNeighs.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.getShNeigh());
        }
        return l;
    }

    /**
     * list rsvp tunnels
     *
     * @param lower forwarder
     * @return list of te tunnels
     */
    public static userFormat rsvpTunnelShow(ipFwd lower) {
        userFormat l = new userFormat("|", "source|id|subgroup|id|target|id|description|");
        for (int i = 0; i < lower.trafEngs.size(); i++) {
            ipFwdTrfng ntry = lower.trafEngs.get(i);
            if (ntry == null) {
                continue;
            }
            l.add("" + ntry);
        }
        return l;
    }

    /**
     * list bfd neighbors
     *
     * @param lower forwarder
     * @return list of neighbors
     */
    public static userFormat bfdNeighShow(ipFwd lower) {
        userFormat l = new userFormat("|", "interface|address|state|uptime|clients");
        for (int i = 0; i < lower.ifaces.size(); i++) {
            ipFwdIface ntry = lower.ifaces.get(i);
            if (ntry == null) {
                continue;
            }
            if (ntry.bfdCfg == null) {
                continue;
            }
            ntry.bfdCfg.getShNeighs(l);
        }
        return l;
    }

    /**
     * list hsrp neighbors
     *
     * @param lower forwarder
     * @return list of neighbors
     */
    public static userFormat hsrpNeighShow(ipFwd lower) {
        userFormat l = new userFormat("|", "interface|address|state|priority");
        for (int i = 0; i < lower.ifaces.size(); i++) {
            ipFwdIface ntry = lower.ifaces.get(i);
            if (ntry == null) {
                continue;
            }
            if (ntry.hsrpCfg == null) {
                continue;
            }
            ntry.hsrpCfg.getShNeighs(l);
        }
        return l;
    }

    /**
     * list vrrp neighbors
     *
     * @param lower forwarder
     * @return list of neighbors
     */
    public static userFormat vrrpNeighShow(ipFwd lower) {
        userFormat l = new userFormat("|", "interface|address|priority");
        for (int i = 0; i < lower.ifaces.size(); i++) {
            ipFwdIface ntry = lower.ifaces.get(i);
            if (ntry == null) {
                continue;
            }
            if (ntry.vrrpCfg == null) {
                continue;
            }
            ntry.vrrpCfg.getShNeighs(l);
        }
        return l;
    }

    /**
     * list pim neighbors
     *
     * @param lower forwarder
     * @return list of neighbors
     */
    public static userFormat pimNeighShow(ipFwd lower) {
        userFormat l = new userFormat("|", "interface|address|priority|uptime");
        for (int i = 0; i < lower.ifaces.size(); i++) {
            ipFwdIface ntry = lower.ifaces.get(i);
            if (ntry == null) {
                continue;
            }
            if (ntry.pimCfg == null) {
                continue;
            }
            ntry.pimCfg.getShNeighs(l);
        }
        return l;
    }

    /**
     * list pim interfaces
     *
     * @param lower forwarder
     * @return list of interfaces
     */
    public static userFormat pimIfaceShow(ipFwd lower) {
        userFormat l = new userFormat("|", "interface|neighbors");
        for (int i = 0; i < lower.ifaces.size(); i++) {
            ipFwdIface ntry = lower.ifaces.get(i);
            if (ntry == null) {
                continue;
            }
            if (ntry.pimCfg == null) {
                continue;
            }
            l.add(ntry + "|" + ntry.pimCfg.neighCount());
        }
        return l;
    }

    /**
     * list routing protocols
     *
     * @param lower forwarder
     * @return list of protocols
     */
    public static userFormat routersShow(ipFwd lower) {
        userFormat res = new userFormat("|", "protocol|process|interfaces|neighbors|computed|advertised");
        for (int o = 0; o < lower.routers.size(); o++) {
            ipRtr rtr = lower.routers.get(o);
            res.add(cfgRtr.num2name(rtr.routerProtoTyp) + "|" + rtr.routerProcNum + "|" + rtr.routerIfaceCount() + "|" + rtr.routerNeighCount() + "|" + rtr.routerComputed.size() + "|" + rtr.routerRedisted.size());
        }
        return res;
    }

    /**
     * get output for show
     *
     * @param lower forwarder
     * @return output
     */
    public static String vrfListShow(ipFwd lower) {
        return lower.ifaces.size() + "|" + lower.actualR.size() + "|" + lower.labeldR.size() + "|" + lower.groups.size() + "|" + lower.trafEngs.size() + "|" + lower.mp2mpLsp.size() + "|" + lower.natTrns.size() + "|" + lower.routers.size() + "|" + lower.cntr.packRx + "|" + lower.cntr.byteRx;
    }

    /**
     * update nat table
     *
     * @param lower forwarder
     */
    protected static synchronized void updateTableNat(ipFwd lower) {
        long tim = bits.getTime();
        for (int i = lower.natTrns.size() - 1; i >= 0; i--) {
            tabNatTraN ntry = lower.natTrns.get(i);
            if (ntry == null) {
                continue;
            }
            if ((tim - ntry.lastUsed) < lower.natTimeout) {
                continue;
            }
            lower.natTrns.del(ntry);
        }
        for (int i = 0; i < lower.natCfg.size(); i++) {
            tabNatCfgN ntry = lower.natCfg.get(i);
            if (ntry == null) {
                continue;
            }
            if (ntry.newSrcIface == null) {
                continue;
            }
            addrIP adr = new addrIP();
            if (lower.ipVersion == ipCor4.protocolVersion) {
                if (ntry.newSrcIface.addr4 != null) {
                    adr.fromIPv4addr(ntry.newSrcIface.addr4);
                }
            } else {
                if (ntry.newSrcIface.addr6 != null) {
                    adr.fromIPv6addr(ntry.newSrcIface.addr6);
                }
            }
            ntry.newSrcAddr = adr;
        }
    }

    /**
     * notify routers about table change
     *
     * @param lower forwarder
     * @param chg main table changed
     */
    protected static synchronized void notifyRouters(ipFwd lower, boolean chg) {
        for (int i = 0; i < lower.routers.size(); i++) {
            ipRtr rtr = lower.routers.get(i);
            if (rtr == null) {
                continue;
            }
            tabRoute<addrIP> tab = new tabRoute<addrIP>("redist");
            for (int o = 0; o < rtr.routerRedisting.size(); o++) {
                ipRtrRed ntry = rtr.routerRedisting.get(o);
                if (ntry == null) {
                    continue;
                }
                ntry.filter(tab, lower.actualR);
            }
            for (int o = 0; o < rtr.routerAdverting.size(); o++) {
                ipRtrAdv ntry = rtr.routerAdverting.get(o);
                if (ntry == null) {
                    continue;
                }
                ntry.filter(tab, lower.actualR);
            }
            boolean diff = tab.differs(rtr.routerRedisted);
            if (chg) {
                rtr.routerOthersChanged();
            }
            if (!diff) {
                continue;
            }
            tab.version = rtr.routerRedisted.version + 1;
            rtr.routerRedisted = tab;
            rtr.routerRedistChanged();
        }
    }

    /**
     * update route table
     *
     * @param lower forwarder
     * @return false if no change, true if updated
     */
    protected static synchronized boolean updateTableRoute(ipFwd lower) {
        tabRoute<addrIP> tabC = new tabRoute<addrIP>("connected");
        tabC.defDist = 0;
        tabC.defMetr = 0;
        tabC.defRouTyp = tabRouteEntry.routeType.conn;
        tabRoute<addrIP> tabL = new tabRoute<addrIP>("routing");
        tabL.defDist = tabRouteEntry.distanMax;
        tabRoute<addrIP> tabA = new tabRoute<addrIP>("locals");
        tabA.defDist = 0;
        tabA.defMetr = 1;
        tabA.defRouTyp = tabRouteEntry.routeType.local;
        for (int i = 0; i < lower.ifaces.size(); i++) {
            ipFwdIface ifc = lower.ifaces.get(i);
            if (!ifc.ready) {
                continue;
            }
            tabRouteEntry<addrIP> prf = tabC.add(3, ifc.network, null);
            prf.iface = ifc;
            prf = tabA.add(3, new addrPrefix<addrIP>(ifc.addr, ifc.addr.maxBits()), null);
            prf.iface = ifc;
            prf.rouTyp = tabRouteEntry.routeType.local;
            addrIP gtw = ifc.gateAddr;
            if (gtw == null) {
                continue;
            }
            prf = tabA.add(3, new addrPrefix<addrIP>(gtw, gtw.maxBits()), null);
            prf.iface = ifc;
            prf.rouTyp = tabRouteEntry.routeType.remote;
            tabListing<tabPrfxlstN, addrIP> pfl = ifc.gatePrfx;
            if (pfl == null) {
                continue;
            }
            for (int o = 0; o < pfl.size(); o++) {
                prf = new tabRouteEntry<addrIP>();
                tabA.updateBase(prf);
                prf.metric = 2;
                prf.prefix = pfl.get(o).getPrefix();
                prf.nextHop = gtw.copyBytes();
                prf.rouTyp = tabRouteEntry.routeType.defpref;
                prf.iface = ifc;
                tabRoute.addUpdatedEntry(2, tabA, prf, ifc.gateRtmp, null, null);
            }
        }
        tabL.mergeFrom(2, tabC, null, true);
        tabL.mergeFrom(2, tabA, null, true);
        for (int i = 0; i < lower.staticR.size(); i++) {
            ipFwdRoute ntry = lower.staticR.get(i);
            if (ntry == null) {
                continue;
            }
            tabRouteEntry<addrIP> imp = ntry.getPrefix();
            if (imp == null) {
                continue;
            }
            if (ntry.iface == null) {
                tabRouteEntry<addrIP> nh = tabC.route(imp.nextHop);
                if (nh == null) {
                    continue;
                }
                imp.iface = nh.iface;
                tabL.add(2, imp, false, true);
            } else {
                ipFwdIface ifc = new ipFwdIface(ntry.iface.ifwNum, null);
                ifc = lower.ifaces.find(ifc);
                if (ifc == null) {
                    continue;
                }
                if (!ifc.ready) {
                    continue;
                }
                if (!ifc.network.matches(imp.nextHop)) {
                    continue;
                }
                imp.iface = ifc;
                tabL.add(2, imp, false, true);
            }
        }
        for (int i = 0; i < lower.routers.size(); i++) {
            ipRtr rtr = lower.routers.get(i);
            if (rtr == null) {
                continue;
            }
            if (rtr.isBGP() != 0) {
                continue;
            }
            tabRoute<addrIP> tabT = rtr.routerComputed;
            tabT.setProto(rtr.routerProtoTyp, rtr.routerProcNum);
            tabL.mergeFrom(2, tabT, null, true);
        }
        tabL.delDistance(tabRouteEntry.distanMax);
        tabA = new tabRoute<addrIP>("locals");
        tabA.mergeFrom(2, tabL, null, true);
        for (int i = 0; i < lower.routers.size(); i++) {
            ipRtr rtr = lower.routers.get(i);
            if (rtr == null) {
                continue;
            }
            if (rtr.isBGP() != 1) {
                continue;
            }
            tabA.mergeFrom(2, rtr.routerComputed, tabL, true);
        }
        for (int i = 0; i < lower.routers.size(); i++) {
            ipRtr rtr = lower.routers.get(i);
            if (rtr == null) {
                continue;
            }
            if (rtr.isBGP() != 2) {
                continue;
            }
            tabA.mergeFrom(2, rtr.routerComputed, null, true);
        }
        tabA.delDistance(tabRouteEntry.distanMax);
        switch (lower.prefixMode) {
            case igp:
                break;
            case all:
                tabL = new tabRoute<addrIP>("labeled");
                tabL.mergeFrom(2, tabA, null, true);
                break;
            default:
                tabL = new tabRoute<addrIP>("labeled");
                break;
        }
        tabRoute.filterTable(tabL, lower.labelFilter);
        for (int i = 0; i < lower.labeldR.size(); i++) {
            tabRouteEntry<addrIP> old = lower.labeldR.get(i);
            if (old == null) {
                continue;
            }
            if (old.labelLoc == null) {
                continue;
            }
            if (old.labelLoc.getValue() == lower.commonLabel.getValue()) {
                continue;
            }
            if (tabLabel.find(old.labelLoc.getValue()) == null) {
                continue;
            }
            tabRouteEntry<addrIP> cur = tabL.find(old);
            if (cur == null) {
                tabLabel.release(old.labelLoc, 2);
                continue;
            }
            cur.labelLoc = old.labelLoc;
        }
        for (int i = 0; i < tabL.size(); i++) {
            tabRouteEntry<addrIP> ntry = tabL.get(i);
            if (ntry.labelLoc != null) {
                continue;
            }
            if (ntry.nextHop == null) {
                ntry.labelLoc = lower.commonLabel;
                continue;
            }
            tabLabelNtry lab = tabLabel.allocate(2);
            ntry.labelLoc = lab;
        }
        for (int i = 0; i < tabA.size(); i++) {
            tabRouteEntry<addrIP> ntry = tabA.get(i);
            tabRouteEntry<addrIP> loc = tabL.find(ntry);
            if (loc != null) {
                ntry.labelLoc = loc.labelLoc;
            }
            if (ntry.nextHop == null) {
                continue;
            }
            if (ntry.labelLoc != null) {
                ntry.labelLoc.setFwdMpls(2, lower, (ipFwdIface) ntry.iface, ntry.nextHop, ntry.labelRem);
            }
            rtrLdpNeigh nei;
            if (ntry.rouTab != null) {
                nei = ntry.rouTab.ldpNeighFind(null, ntry.nextHop, false);
            } else {
                nei = lower.ldpNeighFind(null, ntry.nextHop, false);
            }
            if (nei == null) {
                continue;
            }
            tabRouteEntry<addrIP> rem = null;
            if (ntry.oldHop == null) {
                rem = nei.prefLearn.find(ntry);
            } else {
                tabRouteEntry<addrIP> prf = tabA.route(ntry.oldHop);
                if (prf == null) {
                    continue;
                }
                rem = nei.prefLearn.find(prf);
            }
            if (rem == null) {
                continue;
            }
            ntry.labelRem = tabLabel.prependLabels(ntry.labelRem, rem.labelRem);
            if (loc != null) {
                loc.labelRem = tabLabel.prependLabels(loc.labelRem, rem.labelRem);
            }
            if (ntry.labelLoc != null) {
                ntry.labelLoc.remoteLab = tabLabel.prependLabels(ntry.labelLoc.remoteLab, rem.labelRem);
            }
        }
        lower.commonLabel.setFwdCommon(1, lower);
        if (!tabA.differs(lower.actualR)) {
            return false;
        }
        tabA.version = lower.actualR.version + 1;
        tabL.version = tabA.version;
        tabC.version = tabA.version;
        lower.connedR = tabC;
        lower.actualR = tabA;
        lower.labeldR = tabL;
        return true;
    }

    /**
     * send join to one group
     *
     * @param grp group to join
     * @param need 1=join, 0=prune
     */
    protected static void joinOneGroup(ipFwdMcast grp, int need) {
        if (grp.iface == null) {
            return;
        }
        if (grp.iface.mldpCfg != null) {
            grp.iface.mldpCfg.sendJoin(grp, need == 1);
        }
        if (grp.iface.pimCfg != null) {
            grp.iface.pimCfg.sendJoin(grp, need);
        }
        if (grp.iface.mhostCfg != null) {
            grp.iface.mhostCfg.sendJoin(grp, need == 1);
        }
    }

    /**
     * set multicast source interface
     *
     * @param lower forwarder
     * @param grp group to update
     */
    protected static void updateOneGroup(ipFwd lower, ipFwdMcast grp) {
        grp.iface = null;
        grp.upstream = null;
        tabRouteEntry<addrIP> prf = lower.actualR.route(grp.source);
        if (prf == null) {
            return;
        }
        grp.iface = (ipFwdIface) prf.iface;
        if (prf.nextHop == null) {
            grp.upstream = grp.source.copyBytes();
        } else {
            grp.upstream = prf.nextHop.copyBytes();
        }
    }

    /**
     * update group table
     *
     * @param lower forwarder
     */
    protected static synchronized void updateTableGroup(ipFwd lower) {
        long tim = bits.getTime();
        for (int o = lower.groups.size(); o >= 0; o--) {
            ipFwdMcast grp = lower.groups.get(o);
            if (grp == null) {
                continue;
            }
            updateOneGroup(lower, grp);
            for (int i = grp.flood.size(); i >= 0; i--) {
                ipFwdIface ifc = grp.flood.get(i);
                if (ifc == null) {
                    continue;
                }
                if (ifc.expires < 0) {
                    continue;
                }
                if (ifc.expires > tim) {
                    continue;
                }
                grp.flood.del(ifc);
            }
            if (grp.local) {
                continue;
            }
            if (grp.flood.size() > 0) {
                continue;
            }
            if (grp.label != null) {
                if (lower.mp2mpLsp.find(grp.label) == null) {
                    grp.label = null;
                } else {
                    continue;
                }
            }
            lower.groups.del(grp);
            joinOneGroup(grp, 0);
        }
    }

    /**
     * update echo table
     *
     * @param lower forwarder
     */
    protected static synchronized void updateTableEcho(ipFwd lower) {
        long tim = bits.getTime();
        for (int i = lower.echoes.size(); i >= 0; i--) {
            ipFwdEcho ntry = lower.echoes.get(i);
            if (ntry == null) {
                continue;
            }
            if ((tim - ntry.created) < 10000) {
                continue;
            }
            lower.echoes.del(ntry);
            ntry.notif.wakeup();
        }
    }

    /**
     * update auto mesh tunnels
     *
     * @param lower forwarder
     */
    protected static synchronized void updateTableAutMsh(ipFwd lower) {
        tabRoute<addrIP> tab = new tabRoute<addrIP>("nh");
        for (int i = 0; i < lower.routers.size(); i++) {
            ipRtr rtr = lower.routers.get(i);
            if (rtr == null) {
                continue;
            }
            if (rtr.routerAutoMesh == null) {
                continue;
            }
            rtr.routerNeighList(tab);
        }
        for (int i = lower.autoMesh.size(); i >= 0; i--) {
            clntMplsTeP2p clnt = lower.autoMesh.get(i);
            if (clnt == null) {
                continue;
            }
            tabRouteEntry<addrIP> ntry = new tabRouteEntry<addrIP>();
            ntry.prefix = new addrPrefix<addrIP>(clnt.target, addrIP.size * 8);
            if (tab.find(ntry) != null) {
                if (lower.actualR.route(ntry.prefix.network) != null) {
                    continue;
                }
            }
            if (debugger.clntMplsAutMsh) {
                logger.debug("stopping " + clnt);
            }
            lower.autoMesh.del(clnt);
            clnt.workStop();
        }
        for (int i = 0; i < tab.size(); i++) {
            tabRouteEntry<addrIP> ntry = tab.get(i);
            tabRouteEntry<addrIP> rou = lower.actualR.find(ntry);
            if (rou != null) {
                if (rou.nextHop == null) {
                    continue;
                }
            } else {
                tabRouteEntry<addrIP> old = lower.actualR.route(ntry.prefix.network);
                if (old == null) {
                    continue;
                }
                addrIP hop;
                if (old.rouTyp == tabRouteEntry.routeType.conn) {
                    hop = ntry.prefix.network;
                } else {
                    hop = old.nextHop;
                }
                if (hop == null) {
                    continue;
                }
                rou = new tabRouteEntry<addrIP>();
                rou.prefix = ntry.prefix;
                rou.nextHop = hop;
                rou.iface = old.iface;
                rou.metric = 3;
                rou.rouTyp = tabRouteEntry.routeType.automesh;
                lower.actualR.add(2, rou, false, false);
            }
            clntMplsTeP2p clnt = new clntMplsTeP2p();
            clnt.target = ntry.prefix.network.copyBytes();
            clntMplsTeP2p old = lower.autoMesh.find(clnt);
            if (old != null) {
                ipFwdTrfng trf = old.getTraffEng();
                if (trf == null) {
                    continue;
                }
                if (trf.trgIfc != rou.iface) {
                    continue;
                }
                if (trf.trgHop.compare(trf.trgHop, rou.nextHop) != 0) {
                    continue;
                }
                rou.labelRem = tabLabel.prependLabels(rou.labelRem, tabLabel.int2labels(trf.trgLab));
                if (rou.labelLoc == null) {
                    continue;
                }
                rou.labelLoc.setFwdMpls(2, lower, (ipFwdIface) rou.iface, rou.nextHop, rou.labelRem);
                continue;
            }
            lower.autoMesh.add(clnt);
            clnt.fwdCor = lower;
            clnt.fwdIfc = null;
            clnt.descr = cfgAll.hostName + ":automesh";
            clnt.expr = 0;
            clnt.ttl = 255;
            clnt.prioS = 7;
            clnt.prioH = 7;
            clnt.bndwdt = 0;
            clnt.recRou = false;
            clnt.setUpper(new ifcNull(false, false));
            clnt.workStart();
            if (debugger.clntMplsAutMsh) {
                logger.debug("starting " + clnt);
            }
        }
    }

    /**
     * update traffic engineering table
     *
     * @param lower forwarder
     */
    protected static synchronized void updateTableTrfng(ipFwd lower) {
        long tim = bits.getTime();
        for (int i = lower.trafEngs.size(); i >= 0; i--) {
            ipFwdTrfng ntry = lower.trafEngs.get(i);
            if (ntry == null) {
                continue;
            }
            if (ntry.srcLoc == 1) {
                continue;
            }
            if ((tim - ntry.created) < ntry.timeout) {
                ntry = ntry.getParent();
                ntry = lower.trafEngs.find(ntry);
                if (ntry == null) {
                    continue;
                }
                ntry.created = tim;
                continue;
            }
            lower.trafEngs.del(ntry);
            ntry.labStop();
        }
    }

    /**
     * fill strict first hop
     *
     * @param lower forwarder to use
     * @param pck packet to update
     * @return false on success, true on error
     */
    public static boolean fillRsvpFrst(ipFwd lower, packRsvp pck) {
        tabRouteEntry<addrIP> ntry = lower.actualR.route(pck.getTrg());
        if (ntry == null) {
            return true;
        }
        if (ntry.nextHop == null) {
            return false;
        }
        if (pck.expRout.size() > 0) {
            tabHop hop = pck.expRout.get(0);
            if (hop.strict == true) {
                return false;
            }
        }
        tabHop hop = new tabHop();
        hop.adr = ntry.nextHop.copyBytes();
        hop.strict = true;
        pck.expRout.add(0, hop);
        return false;
    }

    private static packRsvp fillRsvpPack(ipFwd lower, ipFwdTrfng ntry) {
        ntry.trgIfc = findSendingIface(lower, ntry.trgAdr);
        if (ntry.trgIfc == null) {
            return null;
        }
        packRsvp pck = new packRsvp();
        pck.adsBndwdt = ntry.bwdt;
        pck.adsCmtu = ntry.trgIfc.mtu;
        pck.adsHops = 1;
        pck.adsLtncy = 0;
        pck.expRout = new ArrayList<tabHop>();
        tabHop hop = new tabHop();
        hop.adr = ntry.trgAdr.copyBytes();
        hop.strict = false;
        pck.expRout.add(hop);
        pck.flwSpcPcks = ntry.trgIfc.mtu;
        pck.flwSpcPeak = ntry.bwdt;
        pck.flwSpcPlcd = 0;
        pck.flwSpcRate = ntry.bwdt;
        pck.flwSpcSize = 1000;
        pck.hopAdr = ntry.trgIfc.addr.copyBytes();
        pck.hopId = ntry.trgIfc.ifwNum;
        pck.sessAdr = ntry.trgAdr.copyBytes();
        pck.sessId = ntry.trgId;
        pck.subAddr = ntry.trgAdr.copyBytes();
        pck.sessHld = 7;
        pck.sessStp = 7;
        pck.sessFlg = 0x04; // se style
        pck.sessNam = "" + ntry.descr;
        if (ntry.recRou) {
            pck.recRout = new ArrayList<tabHop>();
        }
        pck.sndrAdr = ntry.srcAdr;
        pck.sndrId = ntry.srcId;
        pck.sbgrpOrg = ntry.subAdr;
        pck.sbgrpId = ntry.subId;
        pck.styleVal = 18;
        pck.timeVal = lower.untriggeredRecomputation;
        pck.ttl = 254;
        return pck;
    }

    /**
     * send refresh to local traffic engineering tunnel
     *
     * @param lower forwarder
     * @param ntry tunnel to refresh
     */
    protected static void refreshTrfngAdd(ipFwd lower, ipFwdTrfng ntry) {
        packRsvp pckRrp = fillRsvpPack(lower, ntry);
        if (pckRrp == null) {
            return;
        }
        fillRsvpFrst(lower, pckRrp);
        packHolder pckBin = new packHolder(true, true);
        pckRrp.createHolder(pckBin);
        pckRrp.fillLabReq();
        pckRrp.createDatPatReq(pckBin);
        pckRrp.createHeader(pckBin);
        lower.protoPack(ntry.trgIfc, pckBin);
        if (debugger.rtrRsvpTraf) {
            logger.debug("tx " + pckRrp);
        }
    }

    /**
     * send refresh to local traffic engineering tunnel
     *
     * @param lower forwarder
     * @param ntry tunnel to refresh
     */
    protected static void refreshTrfngDel(ipFwd lower, ipFwdTrfng ntry) {
        packRsvp pckRrp = fillRsvpPack(lower, ntry);
        if (pckRrp == null) {
            return;
        }
        packHolder pckBin = new packHolder(true, true);
        pckBin.clear();
        pckRrp.createHolder(pckBin);
        pckRrp.createDatPatTer(pckBin);
        pckRrp.createHeader(pckBin);
        lower.protoPack(ntry.trgIfc, pckBin);
        if (debugger.rtrRsvpTraf) {
            logger.debug("tx " + pckRrp);
        }
    }

    /**
     * update multipoint ldp table
     *
     * @param lower forwarder
     */
    protected static void updateTableMplsp(ipFwd lower) {
        for (int i = lower.mp2mpLsp.size() - 1; i >= 0; i--) {
            ipFwdMpmp ntry = lower.mp2mpLsp.get(i);
            if (ntry == null) {
                continue;
            }
            ipFwdMcast grp = ipFwdMpmp.decode4multicast(ntry);
            if (grp != null) {
                if (lower.groups.find(grp) == null) {
                    ntry.local = false;
                }
            }
            ntry.updateState(lower);
            if (ntry.local) {
                continue;
            }
            if (grp != null) {
                if (ntry.selfRoot) {
                    if (ntry.neighs.size() > 0) {
                        lower.mcastAddFlood(grp.group, grp.source, ntry);
                        continue;
                    } else {
                        lower.mcastAddFlood(grp.group, grp.source, null);
                    }
                } else {
                    lower.mcastAddFlood(grp.group, grp.source, null);
                }
            }
            if (ntry.neighs.size() > 0) {
                continue;
            }
            lower.mp2mpLsp.del(ntry);
        }
    }

    /**
     * update every table
     *
     * @param lower forwarder
     * @return false if no change, true if updated
     */
    protected static boolean updateEverything(ipFwd lower) {
        boolean chg = updateTableRoute(lower);
        updateTableAutMsh(lower);
        updateTableNat(lower);
        updateTableGroup(lower);
        updateTableEcho(lower);
        updateTableTrfng(lower);
        updateTableMplsp(lower);
        lower.tableChanger();
        return chg;
    }

    /**
     * check for stalled interfaces
     */
    public static void checkVrfs() {
        for (int i = cfgAll.vrfs.size() - 1; i >= 0; i--) {
            cfgVrf ntry = cfgAll.vrfs.get(i);
            ntry.fwd4.hstry.update(ntry.fwd4.cntr);
            ntry.fwd6.hstry.update(ntry.fwd6.cntr);
        }
    }

}
